package net.knarcraft.bookswithoutborders;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;

import net.knarcraft.bookswithoutborders.command.*;
import net.knarcraft.bookswithoutborders.state.EncryptionStyle;
import net.knarcraft.bookswithoutborders.utility.BookFormatter;
import net.knarcraft.bookswithoutborders.utility.EncryptionHelper;
import net.knarcraft.bookswithoutborders.utility.FileHelper;
import net.knarcraft.bookswithoutborders.utility.InputCleaningHelper;
import net.knarcraft.bookswithoutborders.utility.InventoryHelper;
import net.milkbowl.vault.economy.Economy;
import org.bukkit.ChatColor;
import org.bukkit.Material;
import org.bukkit.OfflinePlayer;
import org.bukkit.command.CommandExecutor;
import org.bukkit.command.CommandSender;
import org.bukkit.command.ConsoleCommandSender;
import org.bukkit.command.PluginCommand;
import org.bukkit.command.TabCompleter;
import org.bukkit.configuration.file.FileConfiguration;
import org.bukkit.configuration.file.YamlConfiguration;
import org.bukkit.entity.Player;
import org.bukkit.inventory.ItemFactory;
import org.bukkit.inventory.ItemStack;
import org.bukkit.inventory.PlayerInventory;
import org.bukkit.inventory.meta.BookMeta;
import org.bukkit.plugin.Plugin;
import org.bukkit.plugin.RegisteredServiceProvider;
import org.bukkit.plugin.java.JavaPlugin;

import static net.knarcraft.bookswithoutborders.utility.InputCleaningHelper.cleanString;
import static net.knarcraft.bookswithoutborders.utility.InputCleaningHelper.fixName;

public class BooksWithoutBorders extends JavaPlugin {
    protected static int bookDuplicateLimit;
    //The separating string between the book title and the book author
    public static String titleAuthorSeparator;
    public static String loreSeparator;
    public static final String SLASH = System.getProperty("file.separator");
    protected static List<String> firstBooks;
    protected static String welcomeMessage;
    protected static List<String> existingPlayers;
    public static Economy eco;
    public static Material bookPriceType = null;
    public static double bookPriceQuantity;
    public static boolean authorOnlyCopy;
    protected static boolean useYml;
    public static boolean adminDecrypt;
    protected static ItemFactory itemFactory;
    public static Map<String, List<String>> loadList;
    public static BooksWithoutBorders bwb;
    protected static BooksWithoutBordersListener bL;
    protected static ConsoleCommandSender consoleSender;
    public static String bookFolder;
    public static final ChatColor errorColor = ChatColor.RED;
    public static final ChatColor successColor = ChatColor.GREEN;
    public static final ChatColor commandColor = ChatColor.YELLOW;

    @Override
    public void onEnable() {
        bwb = this;
        consoleSender = this.getServer().getConsoleSender();
        bookFolder = this.getDataFolder().getAbsolutePath() + SLASH + "Books" + SLASH;
        bL = new BooksWithoutBordersListener();
        loadList = new HashMap<>();
        firstBooks = new ArrayList<>();

        if (SLASH != null && init()) {
            this.getServer().getPluginManager().registerEvents(bL, this);
        } else {
            this.getPluginLoader().disablePlugin(this);
        }

        registerCommand("give", new CommandGive(this), new GiveTabCompleter(this));
        registerCommand("decrypt", new CommandDecrypt(this), null);
        registerCommand("groupEncrypt", new CommandGroupEncrypt(this), null);
        registerCommand("delete", new CommandDelete(this), null);
        registerCommand("deletePublic", new CommandDeletePublic(this), null);
        registerCommand("copy", new CommandCopy(this), null);
        registerCommand("unSign", new CommandUnSign(), null);
        registerCommand("encrypt", new CommandEncrypt(this), null);
        registerCommand("setBookPrice", new CommandSetBookPrice(this), null);
        registerCommand("setLore", new CommandSetLore(), null);
        registerCommand("savePublic", new CommandSavePublic(this), null);
        registerCommand("save", new CommandSave(this), null);
        registerCommand("setAuthor", new CommandSetAuthor(), null);
        registerCommand("setTitle", new CommandSetTitle(), null);
        registerCommand("load", new CommandLoad(this), null);
        registerCommand("loadPublic", new CommandLoadPublic(this), null);
        registerCommand("booksWithoutBorders", new CommandBooksWithoutBorders(this), null);
        registerCommand("reload", new CommandReload(this), null);
        registerCommand("givePublic", new CommandGivePublic(this), null);
    }

    /**
     * Registers a command
     *
     * @param commandName <p>The name of the command to register</p>
     * @param executor    <p>The executor to register for the command</p>
     */
    private void registerCommand(String commandName, CommandExecutor executor, TabCompleter tabCompleter) {
        PluginCommand pluginCommand = this.getCommand(commandName);
        if (pluginCommand != null) {
            pluginCommand.setExecutor(executor);
            if (tabCompleter != null) {
                pluginCommand.setTabCompleter(tabCompleter);
            }
        } else {
            sendErrorMessage(consoleSender, "Failed to register command " + commandName);
        }
    }

    protected boolean init() {
        //initializes Item Factory
        try {
            itemFactory = this.getServer().getItemFactory();
        } catch (java.lang.NoSuchMethodError nsmE) {
            sendErrorMessage(consoleSender, "Warning! [BooksWithoutBorders] failed to initialize!");
            sendErrorMessage(consoleSender, "Please confirm the correct version of [BooksWithoutBorders] is");
            sendErrorMessage(consoleSender, "being run for this version of bukkit!");
            return false;
        }

        //Pulls config values
        if (!loadConfig()) {
            return false;
        }

        this.getConfig().set("Options.Save_Books_in_Yaml_Format", useYml);
        this.getConfig().set("Options.Max_Number_of_Duplicates", bookDuplicateLimit);
        this.getConfig().set("Options.Title-Author_Separator", titleAuthorSeparator);
        this.getConfig().set("Options.Lore_line_separator", loreSeparator);
        this.getConfig().set("Options.Books_for_new_players", firstBooks);
        this.getConfig().set("Options.Message_for_new_players", welcomeMessage);
        if (bookPriceType != null) {
            if (bookPriceType != Material.AIR) {
                this.getConfig().set("Options.Price_to_create_book.Item_type", bookPriceType.toString());
            } else {
                this.getConfig().set("Options.Price_to_create_book.Item_type", "Economy");
            }
        } else {
            this.getConfig().set("Options.Price_to_create_book.Item_type", "Item type name");
        }
        this.getConfig().set("Options.Price_to_create_book.Required_quantity", bookPriceQuantity);
        this.getConfig().set("Options.Admin_Auto_Decrypt", adminDecrypt);
        this.getConfig().set("Options.Author_Only_Copy", authorOnlyCopy);

        //Handles old book and quill settings
        if (this.getConfig().contains("Options.Require_book_and_quill_to_create_book")) {
            sendSuccessMessage(consoleSender, "[BooksWithoutBorders] Found old config setting \"Require_book_and_quill_to_create_book\"");
            sendSuccessMessage(consoleSender, "Updating to \"Price_to_create_book\" settings");

            if (this.getConfig().getBoolean("Options.Require_book_and_quill_to_create_book")) {
                bookPriceType = Material.WRITABLE_BOOK;
                bookPriceQuantity = 1;
                this.getConfig().set("Options.Price_to_create_book.Item_type", bookPriceType.toString());
                this.getConfig().set("Options.Price_to_create_book.Required_quantity", bookPriceQuantity);
            }

            this.getConfig().set("Options.Require_book_and_quill_to_create_book", null);
        }

        this.saveConfig();

        File fTest = new File(bookFolder);
        File efTest = new File(bookFolder + "Encrypted" + SLASH);

        if (!fTest.exists()) {
            try {
                if (!fTest.mkdir()) {
                    sendErrorMessage(consoleSender, "Saving failed! Aborting...");
                    return false;
                }
            } catch (Exception e) {
                e.printStackTrace();
                return false;
            }
        }
        if (!efTest.exists()) {
            try {
                if (!efTest.mkdir()) {
                    sendErrorMessage(consoleSender, "Saving failed! Aborting...");
                    return false;
                }
            } catch (Exception e) {
                e.printStackTrace();
                return false;
            }
        }

        return loadExistingPlayers();
    }

    public boolean loadConfig() {
        this.reloadConfig();
        try {
            useYml = this.getConfig().getBoolean("Options.Save_Books_in_Yaml_Format", true);
            bookDuplicateLimit = this.getConfig().getInt("Options.Max_Number_of_Duplicates", 5);
            titleAuthorSeparator = this.getConfig().getString("Options.Title-Author_Separator", ",");
            loreSeparator = this.getConfig().getString("Options.Lore_line_separator", "~");
            adminDecrypt = this.getConfig().getBoolean("Options.Admin_Auto_Decrypt", false);
            authorOnlyCopy = this.getConfig().getBoolean("Options.Author_Only_Copy", false);

            firstBooks = this.getConfig().getStringList("Options.Books_for_new_players");
            if (this.getConfig().contains("Options.Book_for_new_players"))
                firstBooks.add(this.getConfig().getString("Options.Book_for_new_players"));
            if (firstBooks.isEmpty())
                firstBooks.add(" ");

            welcomeMessage = this.getConfig().getString("Options.Message_for_new_players", " ");

            //Converts string into material
            String sMaterial = this.getConfig().getString("Options.Price_to_create_book.Item_type", " ");
            if (sMaterial.equalsIgnoreCase("Economy")) {
                if (setupEconomy())
                    bookPriceType = Material.AIR;
                else {
                    sendErrorMessage(consoleSender, "BooksWithoutBorders failed to hook into Vault! Book price not set!");
                    bookPriceType = null;
                }
            } else if (!sMaterial.equalsIgnoreCase(" ")) {
                for (Material m : Material.values()) {
                    if (m.toString().equalsIgnoreCase(sMaterial))
                        bookPriceType = m;
                }
            }
            bookPriceQuantity = this.getConfig().getDouble("Options.Price_to_create_book.Required_quantity", 0);

            //makes sure titleAuthorSeparator is a valid value
            titleAuthorSeparator = cleanString(titleAuthorSeparator);
            if (titleAuthorSeparator.length() != 1) {
                sendErrorMessage(consoleSender, "Title-Author_Separator is set to an invalid value!");
                sendErrorMessage(consoleSender, "Reverting to default value of \",\"");
                titleAuthorSeparator = ",";
                this.getConfig().set("Options.Title-Author_Separator", titleAuthorSeparator);
            }
        } catch (Exception e) {
            sendErrorMessage(consoleSender, "Warning! Config.yml failed to load!");
            sendErrorMessage(consoleSender, "Try Looking for settings that are missing values!");
            return false;
        }

        return true;
    }

    public boolean setupEconomy() {
        Plugin plugin = getServer().getPluginManager().getPlugin("Vault");

        if (plugin != null) {
            RegisteredServiceProvider<Economy> economyProvider = getServer().getServicesManager().getRegistration(net.milkbowl.vault.economy.Economy.class);
            if (economyProvider != null) {
                eco = economyProvider.getProvider();
            }
        }

        return (eco != null);
    }

    public boolean loadExistingPlayers() {
        File fTest = new File(this.getDataFolder().getAbsolutePath() + SLASH + "Existing Players.txt");
        existingPlayers = new ArrayList<>();

        if (!fTest.exists()) {
            try {
                if (!fTest.createNewFile()) {
                    sendErrorMessage(consoleSender, "Could not create Existing Players file! Aborting...");
                    return false;
                }
                PrintWriter pw = new PrintWriter(new FileWriter(fTest));
                OfflinePlayer[] players = this.getServer().getOfflinePlayers();

                for (OfflinePlayer player : players) {
                    existingPlayers.add(player.getName());
                }
                for (String player : existingPlayers) {
                    pw.println(player);
                }
                pw.close();
            } catch (Exception e) {
                e.printStackTrace();
                return false;
            }
        } else {
            try {
                BufferedReader br = new BufferedReader(new FileReader(fTest));
                String nullCheck = br.readLine();

                while (nullCheck != null) {
                    existingPlayers.add(nullCheck);
                    nullCheck = br.readLine();
                }
                br.close();
            } catch (Exception e) {
                e.printStackTrace();
                return false;
            }
        }

        return true;
    }

    protected void addExistingPlayer(String playerName) {
        existingPlayers.add(playerName);

        try {
            PrintWriter pw = new PrintWriter(new FileWriter(this.getDataFolder().getAbsolutePath() + SLASH + "Existing Players.txt", true));
            pw.println(playerName);
            pw.close();
        } catch (IOException e) {
            sendErrorMessage(consoleSender, "Failed to add " + playerName + " to Existing Players list");
        }
    }

    protected boolean hasPlayedBefore(String playerName) {
        if (!existingPlayers.contains(playerName)) {
            addExistingPlayer(playerName);
            return false;
        }

        return true;
    }

    protected void bookToYml(String path, String fileName, BookMeta book) throws IOException {
        FileConfiguration bookYml = YamlConfiguration.loadConfiguration(new File(path, "blank"));

        if (book.hasTitle())
            bookYml.set("Title", book.getTitle());
        if (book.hasAuthor())
            bookYml.set("Author", book.getAuthor());
        if (book.hasPages())
            bookYml.set("Pages", book.getPages());
        if (book.hasLore())
            bookYml.set("Lore", book.getLore());

        bookYml.save(path + fileName + ".yml");
    }

    protected BookMeta bookFromYml(File file, BookMeta bDat) {
        //Handles old style loading
        if (file.getName().substring(file.getName().lastIndexOf(".")).equals(".txt"))
            bDat = bookFromTXT(file.getName(), file, bDat);
        else {
            try {
                FileConfiguration bookYml = YamlConfiguration.loadConfiguration(file);

                bDat.setTitle(bookYml.getString("Title", "Untitled"));
                bDat.setAuthor(bookYml.getString("Author", "Unknown"));
                bDat.setPages(bookYml.getStringList("Pages"));
                bDat.setLore(bookYml.getStringList("Lore"));
            } catch (IllegalArgumentException e) {
                return null;
            }
        }

        return bDat;
    }

    protected void bookToTXT(String path, String fileName, BookMeta book) throws IOException {
        FileWriter fw = new FileWriter(path + fileName + ".txt");
        PrintWriter pw = new PrintWriter(fw);
        List<String> pages = book.getPages();

        pw.println("[Book]");
        for (String page : pages) {
            pw.println(page);
        }
        pw.close();
    }

    protected BookMeta bookFromTXT(String fileName, File file, BookMeta bDat) {
        String author;
        String title;
        if (fileName.contains(titleAuthorSeparator)) {
            author = fileName.substring(fileName.indexOf(titleAuthorSeparator) + 1, fileName.length() - 4);
            title = fileName.substring(0, fileName.indexOf(titleAuthorSeparator));
        } else {
            author = "Unknown";
            title = fileName.substring(0, fileName.length() - 4);
        }

        title = fixName(title, true);

        List<String> rawPages = new ArrayList<>();
        String nullCheck;

        try {
            FileReader fileReader = new FileReader(file);
            BufferedReader bufferedReader = new BufferedReader(fileReader);

            rawPages.add(bufferedReader.readLine());

            if (rawPages.get(0) == null) {
                bufferedReader.close();
                return null;
            }

            //If a book is being loaded its content need not be adjusted
            if (rawPages.get(0).equalsIgnoreCase("[Book]")) {
                rawPages.remove(0);
                nullCheck = bufferedReader.readLine();
                while (nullCheck != null) {
                    rawPages.add(nullCheck);
                    nullCheck = bufferedReader.readLine();
                }
            } else {
                //Adjusts content to page length
                while (rawPages.get(rawPages.size() - 1) != null) {
                    BookFormatter.formatLastPage(rawPages);
                    rawPages.add(bufferedReader.readLine());
                }
            }
            bufferedReader.close();
        } catch (IOException e) {
            e.printStackTrace();
            return null;
        }

        List<String> pages = new ArrayList<>(InputCleaningHelper.cleanList(rawPages));

        bDat.setAuthor(author);
        bDat.setTitle(title);
        bDat.setPages(pages);

        return bDat;
    }

    /**
     * Saves a book to a file
     *
     * @param player             <p>The player holding the book</p>
     * @param heldBook           <p>The book held</p>
     * @param overwrite          <p>Whether to overwrite any existing books</p>
     * @param saveToPublicFolder <p>Whether to save the book to the public folder instead of the player folder</p>
     */
    public void saveBook(Player player, ItemStack heldBook, boolean overwrite, boolean saveToPublicFolder) {
        BookMeta book = (BookMeta) heldBook.getItemMeta();
        if (book == null) {
            sendErrorMessage(player, "Unable to get metadata for your held book!");
            return;
        }

        String savePath;
        if (saveToPublicFolder) {
            savePath = bookFolder;
        } else {
            savePath = bookFolder + cleanString(player.getName()) + SLASH;
        }

        //Generate book filename
        String fileName;
        if (!book.hasTitle()) {
            fileName = "Untitled," + player.getName();
        } else {
            fileName = book.getTitle() + titleAuthorSeparator + book.getAuthor();
        }
        fileName = cleanString(fileName);
        fileName = fixName(fileName, false);

        //Make sure the used folders exist
        File file = new File(savePath);
        if (!file.exists() && !file.mkdir()) {
            sendErrorMessage(player, "Saving Failed! If this continues to happen, consult server admin!");
            return;
        }
        File[] foundFiles = file.listFiles();
        if (foundFiles == null) {
            sendErrorMessage(player, "Saving Failed! If this continues to happen, consult server admin!");
            return;
        }

        sendErrorMessage(player, String.valueOf(overwrite));
        sendErrorMessage(player, fileName);
        sendErrorMessage(player, Arrays.toString(foundFiles));

        //Find any duplicates of the book
        int foundDuplicates = FileHelper.findDuplicates(foundFiles, fileName);

        //Deal with duplicates
        if (foundDuplicates > 0) {
            //TODO: Decide if this makes sense or needs to be changed
            //Skip duplicate book
            if (!fileName.contains("Untitled") && !overwrite) {
                sendErrorMessage(player, "Book is already saved!");
                sendErrorMessage(player, "Use " + commandColor + "/bwb Save true " + errorColor + "to overwrite!");
                return;
            }

            //Skip if duplicate limit is reached
            if (foundDuplicates > bookDuplicateLimit) {
                sendErrorMessage(player, "Maximum amount of " + fileName + " duplicates reached!");
                sendErrorMessage(player, "Use " + commandColor + "/bwb Save true " + errorColor + "to overwrite!");
                return;
            }

            //Alter duplicated filename
            if (fileName.contains("Untitled") && !overwrite) {
                fileName = "(" + foundDuplicates + ")" + fileName;
            }
        }

        try {
            if (useYml) {
                bookToYml(savePath, fileName, book);
            } else {
                bookToTXT(savePath, fileName, book);
            }

            sendSuccessMessage(player, "Book Saved as \"" + fileName + "\"");
        } catch (IOException e) {
            e.printStackTrace();
        }

    }

    public ItemStack loadBook(CommandSender sender, String fileName, String isSigned, String dir) {
        return loadBook(sender, fileName, isSigned, dir, 1);
    }

    public ItemStack loadBook(CommandSender sender, String fileName, String isSigned, String dir, int numCopies) {
        //checks if player is using list number to load
        for (int x = 0; x < fileName.length(); x++) {
            if (!Character.isDigit(fileName.charAt(x))) {
                break;
            }
            if (x == fileName.length() - 1 && Character.isDigit(fileName.charAt(x)) && loadList.containsKey(sender.getName())) {
                if (Integer.parseInt(fileName) <= loadList.get(sender.getName()).size()) {
                    fileName = loadList.get(sender.getName()).get(Integer.parseInt(fileName) - 1);
                }
            }
        }

        File file;
        if (dir.equalsIgnoreCase("public")) {
            file = FileHelper.getBookFile(bookFolder + fileName);
        } else if (dir.equalsIgnoreCase("player")) {
            file = FileHelper.getBookFile(bookFolder + cleanString(sender.getName()) + SLASH + fileName);
        } else if (dir.trim().isEmpty()) {
            file = FileHelper.getBookFile(bookFolder + "Encrypted" + SLASH + dir + SLASH + fileName);
        } else {
            file = null;
        }
        if (file == null || !file.isFile()) {
            sendErrorMessage(sender, "Incorrect file name!");
            return null;
        }

        ItemStack book;
        BookMeta bookMetadata = (BookMeta) itemFactory.getItemMeta(Material.WRITTEN_BOOK);
        if (isSigned.equalsIgnoreCase("true")) {
            book = new ItemStack(Material.WRITTEN_BOOK);
        } else {
            book = new ItemStack(Material.WRITABLE_BOOK);
        }

        bookFromYml(file, bookMetadata);
        if (bookMetadata == null) {
            sendErrorMessage(sender, "File was blank!!");
            return null;
        }

        if (!dir.equalsIgnoreCase("public") && !dir.equalsIgnoreCase("player") &&
                !dir.equalsIgnoreCase("") && bookMetadata.hasLore()) {
            List<String> oldLore = bookMetadata.getLore();
            if (oldLore != null) {
                List<String> newLore = new ArrayList<>(oldLore);
                newLore.remove(0);
                bookMetadata.setLore(newLore);
            }
        }

        book.setItemMeta(bookMetadata);
        book.setAmount(numCopies);

        if (booksHavePrice() && !sender.hasPermission("bookswithoutborders.bypassBookPrice") &&
                (dir.equalsIgnoreCase("public") || dir.equalsIgnoreCase("player") ||
                        dir.equalsIgnoreCase("")) &&
                cannotPayForBookPrinting((Player) sender, numCopies)) {
            return null;
        }

        return book;
    }

    /**
     * Makes the player pay for printing a given number of books
     *
     * @param player    <p>The player printing the books</p>
     * @param numCopies <p>The number of copies the player is trying to print</p>
     * @return <p>True if the player cannot pay for the printing of the books</p>
     */
    public boolean cannotPayForBookPrinting(Player player, int numCopies) {
        //BookPriceQuantity: How many items are required to pay for each book
        //BookPriceType: Which item is used to pay for the books. AIR = use economy
        double cost = BooksWithoutBorders.bookPriceQuantity * numCopies;
        int itemCost = (int) cost;

        if (BooksWithoutBorders.bookPriceType == Material.AIR) {
            return !payForBookPrintingEconomy(player, cost, numCopies);
        } else {
            if (player.getInventory().contains(BooksWithoutBorders.bookPriceType, itemCost)) {
                payForBookPrintingItem(player, itemCost);
                return false;
            } else {
                BooksWithoutBorders.sendErrorMessage(player, itemCost + " " + BooksWithoutBorders.bookPriceType.toString() +
                        "(s) are required for this command!");
                return true;
            }
        }
    }

    /**
     * Uses economy to take payment for printing a number of books
     *
     * @param player    <p>The player which needs to pay</p>
     * @param cost      <p>The cost of the book printing</p>
     * @param numCopies <p>The number of books the player is printing</p>
     * @return <p>True if the player had the money and it has been withdrawn</p>
     */
    private boolean payForBookPrintingEconomy(Player player, double cost, int numCopies) {
        Economy economy = BooksWithoutBorders.eco;
        if ((economy.getBalance(player) - cost) >= 0) {
            economy.withdrawPlayer(player, cost);
            BooksWithoutBorders.sendSuccessMessage(player, economy.format(cost) + " withdrawn to create " + numCopies + " book(s)");
            BooksWithoutBorders.sendSuccessMessage(player, "New balance: " + economy.format(economy.getBalance(player)));
            return true;
        } else {
            BooksWithoutBorders.sendErrorMessage(player, economy.format(cost) + " is required for this command!");
            return false;
        }
    }

    /**
     * Takes payment for printing a number of books by withdrawing the correct item
     *
     * @param player   <p>The player which needs to pay</p>
     * @param itemCost <p>The number of items to pay</p>
     */
    private void payForBookPrintingItem(Player player, int itemCost) {
        PlayerInventory playerInventory = player.getInventory();

        int clearedAmount = 0;
        while (clearedAmount < itemCost) {
            int firstItemIndex = playerInventory.first(BooksWithoutBorders.bookPriceType);
            ItemStack firstItem = playerInventory.getItem(firstItemIndex);

            if (Objects.requireNonNull(firstItem).getAmount() <= itemCost - clearedAmount) {
                clearedAmount += firstItem.getAmount();
                player.getInventory().clear(firstItemIndex);
            } else {
                clearedAmount = itemCost;
                firstItem.setAmount(firstItem.getAmount() - (clearedAmount));
            }
        }
    }

    /**
     * Saves an encrypted book to be decryptable for the given user
     *
     * @param player       <p>The player encrypting the book</p>
     * @param bookMetaData <p>Metadata for the book to encrypt</p>
     * @param key          <p>The key to use for encryption</p>
     * @return <p>The new encrypted metadata for the book, or null if encryption failed</p>
     */
    protected Boolean saveEncryptedBook(Player player, BookMeta bookMetaData, String key) {
        String path = bookFolder + "Encrypted" + SLASH;
        String fileName = (!bookMetaData.hasTitle()) ? "Untitled," + player.getName() :
                bookMetaData.getTitle() + titleAuthorSeparator + bookMetaData.getAuthor();

        fileName = "[" + key + "]" + fileName;
        fileName = cleanString(fileName);
        fileName = fixName(fileName, false);

        //cancels saving if file is already encrypted
        File file = (useYml) ? new File(path + fileName + ".yml") : new File(path + fileName + ".txt");
        if (file.isFile()) {
            return true;
        }

        try {
            bookToYml(path, fileName, bookMetaData);
        } catch (IOException e) {
            e.printStackTrace();
            sendErrorMessage(player, "Encryption failed!");
            return false;
        }
        return true;
    }

    /**
     * Saves an encrypted book to be decryptable for the given group
     *
     * @param player       <p>The player trying to encrypt the book</p>
     * @param bookMetadata <p>Metadata for the book to encrypt</p>
     * @param groupName    <p>The group which should be able to decrypt the book</p>
     * @return <p>The new encrypted metadata for the book, or null if encryption failed</p>
     */
    protected BookMeta saveEncryptedBookForGroup(Player player, BookMeta bookMetadata, String groupName) {
        String path = bookFolder + "Encrypted" + SLASH + cleanString(groupName) + SLASH;
        File dirTest = new File(path);
        //Creates group dir
        if (!dirTest.exists()) {
            try {
                if (!dirTest.mkdir()) {
                    sendErrorMessage(consoleSender, "Unable to create encryption group folder!");
                    return null;
                }
            } catch (Exception e) {
                e.printStackTrace();
                return null;
            }
        }
        //Creates file
        String fileName = (!bookMetadata.hasTitle()) ? "Untitled," + player.getName() :
                bookMetadata.getTitle() + titleAuthorSeparator + bookMetadata.getAuthor();

        fileName = cleanString(fileName);
        fileName = fixName(fileName, false);

        List<String> newLore = new ArrayList<>();
        newLore.add(ChatColor.GRAY + "[" + groupName + " encrypted]");
        if (bookMetadata.hasLore()) {
            List<String> oldLore = bookMetadata.getLore();
            if (oldLore != null) {
                newLore.addAll(oldLore);
            }
        }
        bookMetadata.setLore(newLore);

        //Save file
        File file = (useYml) ? new File(path + fileName + ".yml") : new File(path + fileName + ".txt");
        if (!file.isFile()) {
            try {
                bookToYml(path, fileName, bookMetadata);
            } catch (IOException e) {
                e.printStackTrace();
                sendErrorMessage(player, "Group encrypted failed!");
                return null;
            }
        }

        return bookMetadata;
    }

    /**
     * Loads an encrypted book
     *
     * @param player              <p>The player trying to load the book</p>
     * @param key                 <p>The encryption key/password for decryption</p>
     * @param deleteEncryptedFile <p>Whether to delete the plaintext file after decryption is finished</p>
     * @return <p>The loaded book, or null if no book could be loaded</p>
     */
    public ItemStack loadEncryptedBook(Player player, String key, boolean deleteEncryptedFile) {
        ItemStack heldBook = InventoryHelper.getHeldBook(player, true);
        BookMeta bookMetadata = (BookMeta) heldBook.getItemMeta();
        String path = bookFolder + "Encrypted" + SLASH;

        if (bookMetadata == null) {
            return null;
        }

        String fileName = (!bookMetadata.hasTitle()) ? "Untitled," + player.getName() : bookMetadata.getTitle() +
                titleAuthorSeparator + bookMetadata.getAuthor();
        fileName = "[" + key + "]" + fileName;
        fileName = cleanString(fileName);
        fileName = fixName(fileName, false);

        File file = new File(path + fileName + ".yml");
        if (!file.isFile()) {
            file = new File(path + fileName + ".txt");

            if (!file.isFile()) {
                sendErrorMessage(player, "Incorrect decryption key!");
                return null;
            }
        } else {
            try {
                bookMetadata = bookFromYml(file, bookMetadata);
            } catch (Exception e) {
                sendErrorMessage(player, "Decryption failed!");
                return null;
            }
        }

        if (deleteEncryptedFile) {
            try {
                if (!file.delete()) {
                    sendErrorMessage(consoleSender, "Book encryption data failed to delete upon decryption!");
                    sendErrorMessage(consoleSender, "File location:" + file.getPath());
                }
            } catch (Exception e) {
                sendErrorMessage(consoleSender, "Book encryption data failed to delete upon decryption!");
                sendErrorMessage(consoleSender, "File location:" + file.getPath());
            }
        }

        ItemStack newBook = new ItemStack(Material.WRITTEN_BOOK);//Book(book.getAuthor(), book.getTitle(), pages, 1, 387);
        newBook.setItemMeta(bookMetadata);
        newBook.setAmount(InventoryHelper.getHeldBook(player, true).getAmount());
        return newBook;
    }

    /**
     * Lists available files
     *
     * @param sender     <p>The command sender looking for files</p>
     * @param listPublic <p>Whether to list public or personal files</p>
     * @param silent     <p>Whether to just return the list without printing it</p>
     * @return <p>A list of available files</p>
     */
    public List<String> listFiles(CommandSender sender, Boolean listPublic, boolean silent) {
        File file;
        if (listPublic) {
            file = new File(bookFolder);
        } else {
            file = new File(bookFolder + cleanString(sender.getName()) + SLASH);
        }
        return FileHelper.listFiles(sender, file, silent);
    }

    /**
     * Encrypts a book
     *
     * @param player   <p>The player encrypting the book</p>
     * @param mainHand <p>Whether the player is holding the book in its main hand</p>
     * @param key      <p>The key/password to use for encryption</p>
     * @param style    <p>The encryption style to use</p>
     * @return <p>An encrypted version of the book</p>
     */
    protected ItemStack encryptBook(Player player, boolean mainHand, String key, EncryptionStyle style) {
        return encryptBook(player, mainHand, key, style, "");
    }

    /**
     * Encrypts a book
     *
     * @param player    <p>The player encrypting the book</p>
     * @param mainHand  <p>Whether the player is holding the book in its main hand</p>
     * @param key       <p>The key/password to use for encryption</p>
     * @param style     <p>The encryption style to use</p>
     * @param groupName <p>The name of the group to encrypt for, or "" otherwise</p>
     * @return <p>An encrypted version of the book</p>
     */
    public ItemStack encryptBook(Player player, boolean mainHand, String key, EncryptionStyle style, String groupName) {
        //converts user supplied key into integer form
        String integerKey = EncryptionHelper.getNumberKeyFromStringKey(key);

        BookMeta book = InventoryHelper.getHeldBookMetadata(player, mainHand);

        if (!book.hasPages()) {
            sendErrorMessage(player, "Book is empty!");
            return null;
        }

        //Save the book's un-encrypted contents to a file
        BookMeta newMetadata = saveBookPlaintext(groupName, player, book, integerKey);
        if (newMetadata == null) {
            return null;
        }

        //Get the encrypted pages
        List<String> encryptedPages = EncryptionHelper.encryptBookPages(book, style, integerKey, player);
        if (encryptedPages == null) {
            return null;
        }

        //Format the last page just in case
        BookFormatter.formatLastPage(encryptedPages);
        //Remove empty pages
        List<String> newPages = InputCleaningHelper.cleanList(encryptedPages);

        ItemStack encryptedBook = createEncryptedBook(book, newPages, player, newMetadata);

        sendSuccessMessage(player, "Book encrypted!");
        return encryptedBook;
    }

    /**
     * Creates a new encrypted book
     *
     * @param book        <p>The book to encrypt</p>
     * @param newPages    <p>The new encrypted pages</p>
     * @param player      <p>The player encrypting the book</p>
     * @param newMetadata <p>The new metadata of the book</p>
     * @return <p>An encrypted version of the book</p>
     */
    private ItemStack createEncryptedBook(BookMeta book, List<String> newPages, Player player, BookMeta newMetadata) {
        //Create the encrypted book
        ItemStack encryptedBook = new ItemStack(Material.WRITTEN_BOOK);
        book.setPages(newPages);
        encryptedBook.setItemMeta(book);

        //Update item amount
        encryptedBook.setAmount(InventoryHelper.getHeldBook(player, true).getAmount());
        //Set new item metadata
        encryptedBook.setItemMeta(newMetadata);

        return encryptedBook;
    }

    /**
     * Saves a book's plain text to a file
     *
     * @param groupName  <p>The group who's allowed to decrypt the book, or ""</p>
     * @param player     <p>The player trying to encrypt the book</p>
     * @param book       <p>The book to encrypt</p>
     * @param integerKey <p>The key used to encrypt the book</p>
     * @return <p>The new metadata for the book, or null if it could not be saved</p>
     */
    private BookMeta saveBookPlaintext(String groupName, Player player, BookMeta book, String integerKey) {
        BookMeta newMetadata = book;
        boolean wasSaved;
        if (groupName.trim().isEmpty()) {
            wasSaved = saveEncryptedBook(player, book, integerKey);
        } else {
            newMetadata = saveEncryptedBookForGroup(player, book, groupName);
            wasSaved = newMetadata != null;
        }
        if (wasSaved) {
            return newMetadata;
        } else {
            return null;
        }
    }

    /**
     * Sends a success message to a command sender (player or a console)
     *
     * @param sender  <p>The sender to send the message to</p>
     * @param message <p>The message to send</p>
     */
    public static void sendSuccessMessage(CommandSender sender, String message) {
        sender.sendMessage(successColor + message);
    }

    /**
     * Sends an error message to a command sender (player or a console)
     *
     * @param sender  <p>The sender to send the message to</p>
     * @param message <p>The message to send</p>
     */
    public static void sendErrorMessage(CommandSender sender, String message) {
        sender.sendMessage(errorColor + message);
    }

    /**
     * Checks whether books have a price for printing them
     *
     * @return <p>True if players need to pay for printing books</p>
     */
    public boolean booksHavePrice() {
        return (bookPriceType != null && bookPriceQuantity > 0);
    }

}
