package net.knarcraft.bookswithoutborders.command;

import net.knarcraft.bookswithoutborders.BooksWithoutBorders;
import net.knarcraft.bookswithoutborders.state.EncryptionStyle;
import net.knarcraft.bookswithoutborders.utility.InventoryHelper;
import org.bukkit.command.Command;
import org.bukkit.command.CommandExecutor;
import org.bukkit.command.CommandSender;
import org.bukkit.entity.Player;
import org.bukkit.inventory.ItemStack;
import org.bukkit.inventory.meta.BookMeta;

import java.util.Objects;

/**
 * Command executor for the encrypt command
 */
public class CommandEncrypt implements CommandExecutor {

    private final BooksWithoutBorders booksWithoutBorders;

    public CommandEncrypt(BooksWithoutBorders booksWithoutBorders) {
        this.booksWithoutBorders = booksWithoutBorders;
    }

    @Override
    public boolean onCommand(CommandSender sender, Command command, String label, String[] args) {
        if (!(sender instanceof Player player)) {
            BooksWithoutBorders.sendErrorMessage(sender, "This command can only be used by a player!");
            return false;
        }

        if (InventoryHelper.notHoldingOneWrittenBookCheck(player, "You must be holding a written book to encrypt it!",
                "You cannot encrypt two books at once!")) {
            return false;
        }

        if (args.length < 1) {
            BooksWithoutBorders.sendErrorMessage(player, "You must specify a key to encrypt a book!");
            return false;
        }
        if (args.length > 2) {
            BooksWithoutBorders.sendErrorMessage(player, "Too many command options specified!");
            return false;
        }

        ItemStack heldBook = InventoryHelper.getHeldBook(player, true);

        if (!((BookMeta) Objects.requireNonNull(heldBook.getItemMeta())).hasPages()) {
            BooksWithoutBorders.sendErrorMessage(player, "Book must have contents to encrypt!");
            return false;
        }

        ItemStack encryptedBook;

        if (args.length == 2) {
            encryptedBook = booksWithoutBorders.encryptBook(player, true, args[0], EncryptionStyle.getFromString(args[1]), "");
        } else {
            encryptedBook = booksWithoutBorders.encryptBook(player, true, args[0], EncryptionStyle.SUBSTITUTION, "");
        }

        if (encryptedBook != null) {
            InventoryHelper.setHeldWrittenBook(player, encryptedBook);
            return true;
        } else {
            return false;
        }
    }
}
