package net.knarcraft.bookswithoutborders.command;

import net.knarcraft.bookswithoutborders.BooksWithoutBorders;
import org.bukkit.command.Command;
import org.bukkit.command.CommandExecutor;
import org.bukkit.command.CommandSender;
import org.bukkit.entity.Player;
import org.bukkit.inventory.ItemStack;

import static net.knarcraft.bookswithoutborders.utility.InputCleaningHelper.cleanString;

/**
 * Command executor for the give public command
 */
public class CommandGivePublic implements CommandExecutor {

    private final BooksWithoutBorders booksWithoutBorders;

    public CommandGivePublic(BooksWithoutBorders booksWithoutBorders) {
        this.booksWithoutBorders = booksWithoutBorders;
    }

    @Override
    public boolean onCommand(CommandSender sender, Command command, String label, String[] args) {
        if (sender instanceof Player) {
            if (!sender.hasPermission("bookswithoutborders.givePublic")) {
                BooksWithoutBorders.sendErrorMessage(sender, " You don't have permission to use this command!");
                return false;
            }
        }

        if (args.length == 2 || args.length > 5) {
            BooksWithoutBorders.sendErrorMessage(sender, "Incorrect number of arguments for this command!");
            return false;
        }

        if (args.length == 1) {
            BooksWithoutBorders.loadList.put(sender.getName(), booksWithoutBorders.listFiles(sender, true, false));
            return true;
        }

        for (int x = 0; x < args[1].length(); x++) {
            if (!Character.isDigit(args[1].charAt(x)))
                break;
            if (x == args[1].length() - 1)
                BooksWithoutBorders.loadList.put(sender.getName(), booksWithoutBorders.listFiles(sender, true, true));
        }

        ItemStack newBook;
        Player receivingPlayer = booksWithoutBorders.getServer().getPlayer(args[2]);
        if (receivingPlayer == null) {
            BooksWithoutBorders.sendErrorMessage(sender, "Player not found!");
            return false;
        }

        if (receivingPlayer.getInventory().firstEmpty() == -1) {
            BooksWithoutBorders.sendErrorMessage(sender, "Receiving player must have space in their inventory to receive books!");
            return false;
        }
        //bwb give [book name] [player] [numCopies] [is signed]
        try {
            if (args.length == 5)
                newBook = booksWithoutBorders.loadBook(sender, cleanString(args[1]), args[4], "public", Integer.parseInt(args[3]));
            else if (args.length == 4) {
                if (args[3].equalsIgnoreCase("true") || args[3].equalsIgnoreCase("false"))
                    newBook = booksWithoutBorders.loadBook(sender, cleanString(args[1]), args[3], "public");
                else
                    newBook = booksWithoutBorders.loadBook(sender, cleanString(args[1]), "true", "public", Integer.parseInt(args[3]));
            } else
                newBook = booksWithoutBorders.loadBook(sender, cleanString(args[1]), "true", "public");

            if (newBook != null) {
                receivingPlayer.getInventory().addItem(newBook);
                BooksWithoutBorders.sendSuccessMessage(sender, "Book sent!");
                BooksWithoutBorders.sendSuccessMessage(receivingPlayer, "Book received!");
                return true;
            } else {
                BooksWithoutBorders.sendErrorMessage(sender, "Book failed to load!");
                return false;
            }
        } catch (NumberFormatException e) {
            BooksWithoutBorders.sendErrorMessage(sender, "Invalid number of book copies specified!");
            return false;
        }
    }

}
