package net.knarcraft.bookswithoutborders.command;

import net.knarcraft.bookswithoutborders.BooksWithoutBorders;
import net.knarcraft.bookswithoutborders.state.EncryptionStyle;
import net.knarcraft.bookswithoutborders.utility.InventoryHelper;
import org.bukkit.command.Command;
import org.bukkit.command.CommandExecutor;
import org.bukkit.command.CommandSender;
import org.bukkit.entity.Player;
import org.bukkit.inventory.ItemStack;
import org.bukkit.inventory.meta.BookMeta;

import java.util.List;

/**
 * Command executor for the group encrypt command
 */
public class CommandGroupEncrypt implements CommandExecutor {

    private final BooksWithoutBorders booksWithoutBorders;

    public CommandGroupEncrypt(BooksWithoutBorders booksWithoutBorders) {
        this.booksWithoutBorders = booksWithoutBorders;
    }

    @Override
    public boolean onCommand(CommandSender sender, Command command, String label, String[] args) {
        if (!(sender instanceof Player player)) {
            BooksWithoutBorders.sendErrorMessage(sender, "This command can only be used by a player!");
            return false;
        }

        if (InventoryHelper.notHoldingOneWrittenBookCheck(player,
                "You must be holding a written book to encrypt it!",
                "You cannot encrypt two books at once!")) {
            return false;
        }

        if (args.length < 2) {
            BooksWithoutBorders.sendErrorMessage(player, "You must specify a group name and key to encrypt a book!");
            return false;
        }
        if (args.length > 3) {
            BooksWithoutBorders.sendErrorMessage(player, "Too many command options specified!");
            return false;
        }

        ItemStack heldBook = InventoryHelper.getHeldBook(player, true);
        BookMeta bookMetadata = (BookMeta) heldBook.getItemMeta();

        if (bookMetadata == null) {
            BooksWithoutBorders.sendErrorMessage(player, "Your book seems to be corrupt!");
            return false;
        }

        if (!bookMetadata.hasPages()) {
            BooksWithoutBorders.sendErrorMessage(player, "Book must have contents to encrypt!");
            return false;
        }

        List<String> lore = bookMetadata.getLore();
        if (bookMetadata.hasLore() && lore != null && lore.get(0).contains(" encrypted]")) {
            BooksWithoutBorders.sendErrorMessage(player, "Book is already group encrypted!");
            return false;
        }

        ItemStack eBook;

        if (args.length == 3) {
            eBook = booksWithoutBorders.encryptBook(player, true, args[1], EncryptionStyle.getFromString(args[2]), args[0]);
        } else {
            eBook = booksWithoutBorders.encryptBook(player, true, args[1], EncryptionStyle.SUBSTITUTION, args[0]);
        }

        if (eBook != null) {
            InventoryHelper.setHeldWrittenBook(player, eBook);
            return true;
        } else {
            return false;
        }
    }

}
