package net.knarcraft.bookswithoutborders.command;

import net.knarcraft.bookswithoutborders.BooksWithoutBorders;
import net.knarcraft.bookswithoutborders.utility.InputCleaningHelper;
import org.bukkit.command.Command;
import org.bukkit.command.CommandExecutor;
import org.bukkit.command.CommandSender;
import org.bukkit.entity.Player;
import org.bukkit.inventory.ItemStack;

/**
 * Command executor for the load command
 */
public class CommandLoad implements CommandExecutor {

    private final BooksWithoutBorders booksWithoutBorders;

    public CommandLoad(BooksWithoutBorders booksWithoutBorders) {
        this.booksWithoutBorders = booksWithoutBorders;
    }

    @Override
    public boolean onCommand(CommandSender sender, Command command, String label, String[] args) {
        return loadBook(sender, args, "player", false);
    }

    /**
     * Loads a stored book
     * @param sender <p>The sender asking to load the book</p>
     * @param args <p>The arguments given</p>
     * @param directory <p>The directory to load from (public/player)</p>
     * @param loadPublic <p>Whether to list public files as loadable</p>
     * @return <p>True if the book was loaded successfully</p>
     */
    boolean loadBook(CommandSender sender, String[] args, String directory, boolean loadPublic) {
        if (!(sender instanceof Player player)) {
            BooksWithoutBorders.sendErrorMessage(sender, "This command can only be used by a player!");
            return false;
        }

        if (player.getInventory().firstEmpty() == -1) {
            BooksWithoutBorders.sendErrorMessage(player, "You must have space in your inventory to load books!");
            return false;
        }

        int argumentCount = args.length;

        //Show books available to the player
        if (argumentCount == 0) {
            BooksWithoutBorders.loadList.put(sender.getName(), booksWithoutBorders.listFiles(player, loadPublic, false));
            return true;
        }

        //Organize and parse input
        String bookIdentifier = args[0];
        String copies = "1";
        String isSigned = "true";
        if (args.length == 3) {
            copies = args[1];
            isSigned = args[2];
        } else if (args.length == 2) {
            if (args[1].equalsIgnoreCase("true") || args[1].equalsIgnoreCase("false")) {
                isSigned = args[1];
            } else {
                copies = args[1];
            }
        }

        //Load books available to the player
        try {
            Integer.parseInt(bookIdentifier);
            BooksWithoutBorders.loadList.put(player.getName(), booksWithoutBorders.listFiles(player, loadPublic, true));
        } catch (NumberFormatException ignored) {}

        String bookToLoad = InputCleaningHelper.cleanString(bookIdentifier);
        try {
            //Give the new book if it cannot be loaded
            ItemStack newBook = booksWithoutBorders.loadBook(player, bookToLoad, isSigned, directory, Integer.parseInt(copies));
            if (newBook != null) {
                player.getInventory().addItem(newBook);
                BooksWithoutBorders.sendSuccessMessage(player, "Book created!");
                return true;
            } else {
                BooksWithoutBorders.sendErrorMessage(player, "Book failed to load!");
                return false;
            }
        } catch (NumberFormatException e) {
            BooksWithoutBorders.sendErrorMessage(player, "Invalid number of book copies specified!");
            return false;
        }
    }

}
